package hr.pmf.abkloubcar.drd;

import com.google.gson.GsonBuilder;
import hr.pmf.abkloubcar.drd.model.InputParameters;
import hr.pmf.abkloubcar.drd.model.Node;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.List;
import java.util.stream.Collectors;

import static hr.pmf.abkloubcar.drd.util.Util.*;

public class DRDSolver {


    public static String PARAMETERS = "parameters.json";


    public static List<Node> loadGraph(File file) throws IOException {
        List<String> lines = Files.readAllLines(file.toPath()).stream().filter(line -> line.trim().length() > 0).collect(Collectors.toList());

        String infoLine = lines.get(0);
        String[] parts = infoLine.split("\\s+");
        var verticesNum = Integer.parseInt(parts[0]);
        var edgesNum = Integer.parseInt(parts[1]);

        List<Node> graph = new ArrayList<>();
        for (var verticesIndex = 0; verticesIndex < verticesNum; verticesIndex++) {
            graph.add(new Node(verticesIndex));
        }

        lines.subList(1, edgesNum + 1).forEach(edgeLine -> {
            String[] edgeParts = edgeLine.split("\\s+");
            var nodeA = Integer.parseInt(edgeParts[1]);
            var nodeB = Integer.parseInt(edgeParts[2]);

            graph.get(nodeA).neighbours.add(graph.get(nodeB));
            graph.get(nodeB).neighbours.add(graph.get(nodeA));
        });

        String costLine = lines.get(lines.size() - 1);
        String[] costs = costLine.split("\\s+");
        for (var costIndex = 0; costIndex < costs.length; costIndex++) {
            graph.get(costIndex).cost = Integer.parseInt(costs[costIndex]);
        }
        return graph;
    }

    public static void main(String[] args) throws IOException {

        var gson = new GsonBuilder().setPrettyPrinting().create();
        InputParameters inputParameters = gson.fromJson(Files.readString(Paths.get(RESOURCES, PARAMETERS)), InputParameters.class);

        List<InputParameters> combinations = new ArrayList<>();
        if (!inputParameters.allCombinations) {
            combinations.add(inputParameters);
        } else {
            Arrays.stream(DominatingSetSolver.DominatingSetSelection.values()).forEach(selection -> {
                Arrays.stream(DominatingSetSolver.DominatingSetSort.values()).forEach(sort -> {
                    Arrays.stream(InitialWeightsSolver.InitialWeightsVariant.values()).forEach(variant -> {
                        combinations.add(new InputParameters(inputParameters.logLevel, inputParameters.append, selection, sort, variant, false));
                    });
                });
            });
        }
        log("START");



        var INPUT_FOLDER = "input";
        var OUTPUT_FOLDER = "output";
        var CPLEX_FOLDER = "cplex";

        var OUTPUT_FILE = "result.csv";

        createFileIfNotExisting(OUTPUT_FOLDER, OUTPUT_FILE);
        if (!inputParameters.isAppend()) {
            resetFile(OUTPUT_FOLDER, OUTPUT_FILE);
        }

        for (var parameters : combinations) {


            LOG_LEVEL = parameters.getLogLevel();

            log("####################");
            log("Using " + parameters);
            log("--------------------");

            var files = Paths.get(RESOURCES, INPUT_FOLDER).toFile().listFiles();

            assert files != null;
            for (var file : files) {
                if (file.isDirectory()) {
                    continue;
                }
                log("Processing " + file.getName());

                var graph = loadGraph(file);
                var startTime = System.currentTimeMillis();
                DominatingSetSolver.process(graph, parameters.dominatingSetSelection, parameters.dominatingSetSort);
                InitialWeightsSolver.process(graph, parameters.initialWeightsVariant);
                var preparationTime = System.currentTimeMillis() - startTime;
                var initialCost = graph.stream().map(x -> x.weight * x.cost).reduce(0, Integer::sum);
                //graph.sort(Comparator.comparing(Node::getCost));
                ImprovementSolver.improve(graph);

                var totalTime = System.currentTimeMillis() - startTime;
                if(totalTime < 1) {
                    totalTime = 1;
                }
                graph.sort(Comparator.comparing(Node::getIndex));

                var endCost = graph.stream().map(x -> x.weight * x.cost).reduce(0, Integer::sum);

                List<String> cplexLines = Files.readAllLines(Paths.get(RESOURCES, CPLEX_FOLDER, "solution_" + file.getName())).stream().filter(line -> line.trim().length() > 0).collect(Collectors.toList());
                var cplexTime = Float.parseFloat(cplexLines.get(0).split(": ")[1].split(" CPU")[0]) * 1000;
                var cplexCost = Integer.parseInt(cplexLines.get(2));

                var output = getCsv(file.getName(), preparationTime, totalTime, initialCost, endCost, cplexTime, cplexCost);



                Files.write(Paths.get(RESOURCES, OUTPUT_FOLDER, OUTPUT_FILE), output.getBytes(), StandardOpenOption.APPEND);

                log("Time " + (System.currentTimeMillis() - startTime));
                log("Processed total cost " + graph.stream().filter(x -> x.weight > 0).map(x -> x.index + " " + x.weight).collect(Collectors.joining(" | ")));
                log("Processed total cost " + graph.stream().map(x -> x.weight * x.cost).reduce(0, Integer::sum));
                log("Processed " + file.getName());
                log("--------------------");

            }
        }

        log("END");
    }
}
